/*==================================================================================
| code\stack256.v: stacks for the sp2 cpu                                          |
|                                                                                  |
|         Copyright 2000-2007 by John Rible, <john@sandpipers.com>                 |
|  Permission is granted for individual non-commercial use, provided this notice   |
|  is included. Contact the author concerning any other use.                       |
|                                                                                  |
|  My Verilog name conventions:  - combinatorial signals have lower-case names     |
|     80 lines per page, 2-up    - registers and flipflops have UPPERCASE NAMES    |
|     84 characters per line     - functions and tasks have Capitalized Names      |
====================================================================================

   The stack is implemented with a 256x18 RAM block. The stack pointer is
   implemented with an up/down counter. The interface is generic.

------------------------------------------------------------------------------------
20001116 jwr created this file.
20020822 jwr simplified the LFSR pointer
20070716 jwr BEI edits: larger stack, no TOS, and redo pointer
20070721 jwr drop to 16 bits, clean up code
20070803 jwr move to rising edge, resolve 'r/w at same address' issue
==================================================================================*/

`timescale 1 ns/100 ps

module stack( clock, reset, pop, push, enable, push_bus, pop_bus);

   input         clock;    wire        clock;
   input         reset;    wire        reset;
                                               // stack access: idle pop push error
   input         pop;      wire        pop;    //           re:   0   1    0    1
   input         push;     wire        push;   //           we:   0   0    1    1

   input         enable;   wire        enable;
   input  [15:0] push_bus; wire [15:0] push_bus;
   output [15:0] pop_bus;  wire [15:0] pop_bus, top_bus;

   reg [15:0] NEW_TOP;
   reg  [8:0] decRPTR, decWPTR;
   reg  [7:0] WPTR, newWPTR, RPTR, newRPTR;
   reg        pushed, popped, SAME;

   always @* begin: active

     pushed = push & ~reset & enable;
     popped = pop  & ~reset & enable;
     decRPTR = RPTR - 8'b1;
     decWPTR = WPTR + 8'b1;
     newRPTR = pushed? WPTR: popped? decRPTR[7:0]: RPTR; // pops decrement both
     newWPTR = popped? RPTR: pushed? decWPTR[7:0]: WPTR; // pushes increment both

   end // active

   always @(posedge clock) begin: ptr

      if (reset) begin
          WPTR <= 8'h00;
          RPTR <= 8'hff;
          SAME <= 1'b1;
          NEW_TOP <= 16'b0;
      end
      else begin
          WPTR <= newWPTR; // pushes increment *PTR
          RPTR <= newRPTR; // pops decrement *PTR
          SAME <= pushed;
          NEW_TOP <= push_bus;
      end
   end // ptr

// instantiate ram -----------------------------------------------------------------
// two cycle synchronous, 16-bit word-addressed local memory blocks

   ram256x16 s( .CLK(clock), .WE(pushed), .WA(WPTR), .MI(push_bus),
                                          .RA(RPTR), .MO(top_bus));

   assign pop_bus = SAME? NEW_TOP: top_bus;

endmodule // stack =================================================================
